#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib
import matplotlib.pyplot as plt

import os
import sys
import copy
import time

import math
import numpy as np
import threading as thread

sys.path.append('../')
sys.path.append('../redcp/hexapod_robot')

from redcp import * # import data types

import HexapodRobot 

from ControllerReactive import Controller

#from Explorer import Explorer

from Frontiers import Frontiers
from Planner import Planner
from Mapper import Mapper

CELL_FREE = 0.4 ## < value of free cell
CELL_OCCU = 0.6 ## > value of occupied cell
DELTA_DISTANCE = 0.6 ## more, since goal is exploration

ROBOT_SIZE = 0.5 ## larger than 0.6 has trouble planning path (too much growth) 
WAYPOINT_REACH_DISTANCE = 0.5 # might be similar to robot size

class ExplorerBase:
    def __init__(self):
        """Initialize class
        """
        ## occupancy grid map of the robot ... possibly extended initialization needed in case of 'm1' assignment
        self.gridmap = OccupancyGrid()
        self.gridmap.resolution = 0.1

        self.gridmap.width = 100 # It should be sufficiently large
        self.gridmap.height = 100
        self.gridmap.origin = Pose(Vector3(-5.0,-5.0,0.0), Quaternion(1,0,0,0))
        self.gridmap.data = 0.5 * np.ones((self.gridmap.height * self.gridmap.width))

        ## mutual information grid
        self.grid_mi = None
        
        ## traveled path
        self.path = Path()
        self.planned_path = None
        self.simplified_planned_path = None
        self.frontiers = None
        self.nav_gridmap = None

        self.waypoints = None
        self.new_plan = False
 
        ## stopping condition
        self.stop = False
 
        """Connecting the simulator
        """
        self.robot = HexapodRobot.HexapodRobot(Controller())
        self.mapper = Mapper()
        self.frontiers_extractor = Frontiers()
        self.planner = Planner()
 
        ## sleepy times for the threads
        self.SLEEP_MAPPING     = 1. / 5.0 # ~  5 Hz mapping
        self.SLEEP_PLANNING    = 1. / 0.2 # ~ 0.2 Hz planning once per 5 seconds
        self.SLEEP_PATH_FOLLOW = 1. / 5.0 # ~  5 Hz path following
        
        ## mutexes for accessing gridmap and data of the path_following and planning threads
        self.mutex_gridmap = thread.Lock()
        self.mutex_path_following = thread.Lock()
        self.mutex_planning = thread.Lock()
        self.mutex_stop = thread.Lock()

    def start(self):
        """ method to connect to the simulated robot and start the navigation, localization, mapping and planning
        """
        self.robot.turn_on() #turn on the robot 
        print("[Explorer] Start robot navigation thread")
        self.robot.start_navigation() #start navigation thread
        print("[Explorer] Wait for the robot to be turned on")
        time.sleep(3)
        print("[Explorer] Starting threads")
 
        try: ## start the mapping thread
            mapping_thread = thread.Thread(target = self.mapping)
            mapping_thread.start() 
        except:
            print("Error: unable to start mapping thread")
            sys.exit(1)
        print("[Explorer] Starting mapping thread")
 
        ## start planning thread
        try:
            planning_thread = thread.Thread(target = self.planning)
            planning_thread.start() 
        except:
            print("Error: unable to start planning thread")
            sys.exit(1)
        print("[Explorer] Starting planning thread")
        
        ## start path following
        try:
            path_follow_thread = thread.Thread(target = self.path_following)
            path_follow_thread.start() 
        except:
            print("Error: unable to start planning thread")
            sys.exit(1)
        print("[Explorer] Starting path following thread")
        print("[Explorer] Start end")
        return 

    def is_stopped(self):
        self.mutex_stop.acquire()
        stop = self.stop
        self.mutex_stop.release()
        return stop

    def __del__(self):
        ## turn off the robot
        self.robot.stop_navigation()
        self.robot.turn_off()
 
